import os

__all__ = ['load']


def load(filename: str):
    with open(filename, 'r', encoding='utf8') as f:
        for line in f:
            if '=' not in line:
                continue

            try:
                line = line[:line.index('#')]
            except ValueError:
                pass
            line = line.strip()
            key, value = line.split('=', 1)
            if '{' in value:
                try:
                    value = value.format(**os.environ)
                except KeyError:
                    pass
            os.environ[key] = value



import asyncio
from queue import SimpleQueue

import aioredis
from channels_redis.core import RedisChannelLayer, ConnectionPool, _wrap_close

__all__ = ['QueueConnectionPool', 'AutoReconnectRedisChannelLayer']


class QueueConnectionPool(ConnectionPool):
    def _ensure_loop(self, loop):
        if loop is None:
            loop = asyncio.get_event_loop()

        if loop not in self.conn_map:
            _wrap_close(loop, self)
            self.conn_map[loop] = SimpleQueue()

        return self.conn_map[loop], loop

    async def pop(self, loop=None):
        conns, loop = self._ensure_loop(loop)
        while not conns.empty():
            conn = conns.get()
            try:
                if not conn.closed:
                    break
            except Exception:
                conns.put(conn)
                raise
        else:
            conn = await aioredis.create_redis(**self.host, loop=loop)
        self.in_use[conn] = loop
        return conn

    def push(self, conn):
        loop = self.in_use[conn]
        del self.in_use[conn]
        if loop is not None:
            conns, _ = self._ensure_loop(loop)
            conns.put(conn)

    async def close_loop(self, loop):
        if loop in self.conn_map:
            conns = self.conn_map[loop]
            while not conns.empty():
                conn = conns.get()
                try:
                    conn.close()
                    await conn.wait_closed()
                except Exception:
                    conns.put(conn)
                    raise
            del self.conn_map[loop]

        for k, v in self.in_use.items():
            if v is loop:
                self.in_use[k] = None
                break

    async def close(self):
        conn_map = self.conn_map
        in_use = self.in_use
        self.reset()
        for conns in conn_map.values():
            while not conns.empty():
                conn = conns.get()
                try:
                    conn.close()
                    await conn.wait_closed()
                except Exception:
                    conns.put(conn)
                    raise
        for conn in in_use:
            conn.close()
            await conn.wait_closed()


class AutoReconnectRedisChannelLayer(RedisChannelLayer):
    def __init__(
        self,
        hosts=None,
        prefix="asgi:",
        expiry=60,
        group_expiry=86400,
        capacity=100,
        channel_capacity=None,
        symmetric_encryption_keys=None,
    ):
        super().__init__(hosts, prefix, expiry, group_expiry, capacity, channel_capacity, symmetric_encryption_keys)
        self.pools = [QueueConnectionPool(host) for host in self.hosts]



import random
from typing import Type, Union, Optional

from django.db.models import Model, QuerySet

__all__ = ['random_instance']


def random_instance(query_set: Union[Type[Model], QuerySet])->Optional[Model]:
    if type(query_set) is not QuerySet:
        query_set = query_set.objects.all()

    cnt = query_set.count()
    if not cnt:
        return None

    return query_set[random.randrange(cnt)]



from typing import Union, Optional

from django.http import HttpRequest
from rest_framework.request import Request

from rest_base.settings import base_settings

__all__ = ['get_client_ip']


def get_client_ip(request: Union[HttpRequest, Request]) -> Optional[str]:
    meta = request.META

    for key in base_settings.IP_HEADERS:
        ip: str = meta.get(key)
        if ip:
            if ',' in ip:
                return ip.split(',')[0].strip()
            return ip



from typing import Dict, List, TypeVar, Set, Hashable

__all__ = ['get_dict_list', 'get_dict_set', 'get_dict_dict']

Key = TypeVar('Key', bound=Hashable)
Value = TypeVar('Value')


def get_dict_list(d: Dict[Key, List[Value]], key: Key) -> List[Value]:
    _list = d.get(key)
    if _list is None:
        _list = d[key] = list()
    return _list


def get_dict_set(d: Dict[Key, Set[Value]], key: Key) -> Set[Value]:
    _set = d.get(key)
    if _set is None:
        _set = d[key] = set()
    return _set


def get_dict_dict(d: Dict[Key, Dict[Value]], key: Key) -> Dict[Value]:
    _dict = d.get(key)
    if _dict is None:
        _dict = d[key] = dict()
    return _dict



import os
import time
from argparse import ArgumentParser

from django.conf import settings
from django.core.management import call_command

from rest_base.commands import BaseCommand
from rest_base.settings import base_settings


class Command(BaseCommand):
    help = (
        'Load predefined model instances'
    )

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument('model', type=str, help='Specifies the model to load in the format of app_label.ModelName')
        parser.add_argument(
            '-f', '--filename', nargs='?', type=str, help='Specifies the file name of dumps (default: ModelName.json)')

    def handle(self, *args, **options):
        model: str = options['model']
        filename: str = options['filename']
        if filename is None:
            filename = f"{model.split('.')[-1]}.json"

        try:
            base_dir = settings.BASE_DIR
        except AttributeError as e:
            raise AttributeError('BASE_DIR must be defined in Django settings') from e
        path = os.path.join(base_dir, base_settings.PREDEFINED_ROOT, filename)

        t = time.time()
        self.log(f'load {model} instances from:')
        self.log(path)

        call_command('loaddata', path)

        self.log(f'done ({time.time() - t:.2f} s)')



from typing import Callable, Union, Iterable, Type, Sequence, Optional

from rest_framework.authentication import BaseAuthentication
from rest_framework.decorators import api_view, permission_classes, authentication_classes
from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.views import APIView

__all__ = ['method_branch']


def _perform_authentication(self, request: Request):
    pass


APIView.perform_authentication = _perform_authentication


def _authenticate(authenticators: Union[None, Iterable[Type[BaseAuthentication]]], request: Request):
    if not authenticators:
        return

    for authenticator in authenticators:
        user_auth_tuple = authenticator().authenticate(request)
        if user_auth_tuple is not None:
            return user_auth_tuple


def _authentication_header(
        authentications: Union[None, Sequence[Type[BaseAuthentication]]], request: Request
) -> Optional[str]:
    if authentications:
        return authentications[0]().authenticate_header(request)


def _check_permissions(
        permissions: Union[None, Iterable[Type[BasePermission]]], request: Request, view: APIView
) -> bool:
    if not permissions:
        return True

    for permission in permissions:
        if not permission().has_permission(request, view):
            return False

    return True


def method_branch(
        *,
        GET: Callable[..., Response] = None,
        POST: Callable[..., Response] = None,
        PUT: Callable[..., Response] = None,
        DELETE: Callable[..., Response] = None,
):
    methods = list()

    default_authentications = api_settings.DEFAULT_AUTHENTICATION_CLASSES
    default_permissions = api_settings.DEFAULT_PERMISSION_CLASSES

    get_authentications = post_authentications = put_authentications = delete_authentications = None
    get_permissions = post_permissions = put_permissions = delete_permissions = None
    if GET is not None:
        methods.append('GET')
        get_authentications = getattr(GET, 'authentication_classes', default_authentications)
        get_permissions = getattr(GET, 'permission_classes', default_permissions)
    if POST is not None:
        methods.append('POST')
        post_authentications = getattr(POST, 'authentication_classes', default_authentications)
        post_permissions = getattr(POST, 'permission_classes', default_permissions)
    if PUT is not None:
        methods.append('PUT')
        put_authentications = getattr(PUT, 'authentication_classes', default_authentications)
        put_permissions = getattr(PUT, 'permission_classes', default_permissions)
    if DELETE is not None:
        methods.append('DELETE')
        delete_authentications = getattr(DELETE, 'authentication_classes', default_authentications)
        delete_permissions = getattr(DELETE, 'permission_classes', default_permissions)

    class AuthenticationClass(BaseAuthentication):
        def authenticate(self, request: Request):
            if request.method == 'GET':
                return _authenticate(get_authentications, request)
            if request.method == 'POST':
                return _authenticate(post_authentications, request)
            if request.method == 'PUT':
                return _authenticate(put_authentications, request)
            if request.method == 'DELETE':
                return _authenticate(delete_authentications, request)

        def authenticate_header(self, request: Request) -> Optional[str]:
            if request.method == 'GET':
                return _authentication_header(get_authentications, request)
            if request.method == 'POST':
                return _authentication_header(post_authentications, request)
            if request.method == 'PUT':
                return _authentication_header(put_authentications, request)
            if request.method == 'DELETE':
                return _authentication_header(delete_authentications, request)

    class PermissionClass(BasePermission):
        def has_permission(self, request: Request, view: APIView):
            if request.method == 'GET':
                return _check_permissions(get_permissions, request, view)
            if request.method == 'POST':
                return _check_permissions(post_permissions, request, view)
            if request.method == 'PUT':
                return _check_permissions(put_permissions, request, view)
            if request.method == 'DELETE':
                return _check_permissions(delete_permissions, request, view)
            return True

    @api_view(methods)
    @permission_classes((PermissionClass,))
    @authentication_classes((AuthenticationClass,))
    def branch(request: Request, *args, **kwargs):
        if request.method == 'GET':
            return GET(request, *args, **kwargs)
        if request.method == 'POST':
            return POST(request, *args, **kwargs)
        if request.method == 'PUT':
            return PUT(request, *args, **kwargs)
        if request.method == 'DELETE':
            return DELETE(request, *args, **kwargs)

    return branch



from django.core.cache import cache
from django.utils import timezone

__all__ = ['is_cache_token_left']


def is_cache_token_left(token_id: str, token_max_cnt: int, token_duration: timezone.timedelta) -> bool:
    token = cache.get(token_id)
    if token is None:
        token_cnt = token_max_cnt
        token_cooldown = timezone.now() + token_duration
    else:
        token_cnt, token_cooldown = token
        if token_cooldown < timezone.now():
            token_cnt = token_max_cnt
            token_cooldown = timezone.now() + token_duration

    if token_cnt <= 0:
        return False
    token_cnt -= 1

    cache.set(token_id, (token_cnt, token_cooldown), token_duration.total_seconds())

    return True



import inspect
import multiprocessing
import os
import random
import traceback
from typing import Callable, Union, Iterable, List, Tuple, Set, Optional

from django import db
from django.db.models import QuerySet

__all__ = ['WorkerPool']

WorkerTask = Set[list]
Task = Tuple[Callable, Optional[List[WorkerTask]], Optional[int], tuple, dict, str]


class WorkerPool:
    def __init__(
            self, worker_cnt: int, blocking: bool = True, log: Callable = None,
            verbose: bool = False, throw: bool = True
    ):
        self._pid = None
        self._log = log or print
        self._verbose = verbose
        self._throw = throw
        self.log(f'WorkerPool (size: {worker_cnt}) generated', debug=True)

        self._tasks: List[Task] = list()
        self._worker_cnt = worker_cnt
        self._blocking = blocking

    def __len__(self):
        return len(self._tasks)

    def __bool__(self):
        return bool(self._tasks)

    def _set_pid(self, pid: int):
        self._pid = pid

    def log(self, *args, debug: bool = False):
        if debug and not self._verbose:
            return
        if self._pid is None:
            self._log('[parent]', *args)
        else:
            self._log(f'[{self._pid}]', *args)

    def push(
            self, queryset_or_objects: Union[QuerySet, Iterable], func: Callable,
            *args, plural_name: str = None, **kwargs
    ):
        if type(queryset_or_objects) is QuerySet:
            objects = list(queryset_or_objects.order_by('?'))
            plural_name = plural_name or queryset_or_objects.model._meta.verbose_name_plural
        else:
            objects = random.sample(queryset_or_objects, k=len(queryset_or_objects))
            plural_name = plural_name or f'{func.__name__}s'

        if not objects:
            return

        task_size = (len(objects) + self._worker_cnt - 1) // self._worker_cnt
        worker_tasks: List[WorkerTask] = list()

        for w in range(self._worker_cnt):
            worker_tasks.append(set(objects[w * task_size:(w + 1) * task_size]))

        self._tasks.append((func, worker_tasks, task_size, args, kwargs, plural_name))
        self.log(f'distribute {task_size} / {len(objects)} {plural_name} (func: {func.__name__})', debug=True)

    def register(self, func: Callable, *args, func_name: str = None, **kwargs):
        func_name = func_name or func.__name__
        self._tasks.append((func, None, None, args, kwargs, func_name))
        self.log(f'register {func_name}', debug=True)

    def run(self, set_pid: Callable = None):
        if not self._tasks:
            self.log('no tasks to process', debug=True)
            return

        if set_pid is None:
            set_pid = self._set_pid

        self.log('spawning workers', debug=True)

        workers_pid = list()
        worker_task = [multiprocessing.JoinableQueue() for _ in range(self._worker_cnt)]
        db.connections.close_all()
        for w in range(self._worker_cnt):
            pid = os.fork()
            if pid:
                workers_pid.append(pid)
                continue
            try:
                set_pid(os.getpid())
                db.connection.connect()
                self.log('worker spawned', debug=True)

                for func, worker_tasks, task_size, args, kwargs, name in self._tasks:
                    if self._blocking:
                        worker_task[w].get()
                    try:
                        try:
                            if 'log' in inspect.signature(func).parameters.keys():
                                def func_log(*_args, **_kwargs):
                                    func(*_args, **_kwargs, log=self.log)
                            else:
                                func_log = func
                        except ValueError:
                            func_log = func

                        if worker_tasks is None:
                            self.log(f'process {name}')
                            func_log(*args, **kwargs)
                        else:
                            if worker_tasks[w]:
                                self.log(
                                    f'process {name}: {w * task_size} {w * task_size + len(worker_tasks[w])}',
                                    debug=True,
                                )
                                func_log(worker_tasks[w], *args, **kwargs)
                            else:
                                self.log(f'process {name}: none', debug=True)
                    except:
                        if self._throw:
                            raise
                        self.log('task failed')
                        self.log(traceback.format_exc())
                    finally:
                        if self._blocking:
                            worker_task[w].task_done()

                db.connections.close_all()
            except:
                self.log('worker failed')
                self.log(traceback.format_exc())
            finally:
                os._exit(0)

        if self._blocking:
            for _ in range(len(self._tasks)):
                for w in range(self._worker_cnt):
                    worker_task[w].put(None)
                for w in range(self._worker_cnt):
                    worker_task[w].join()
                self.log('task_done -> next_task', debug=True)

        for pid in workers_pid:
            os.waitpid(pid, 0)

        db.connection.connect()

        self._tasks.clear()
        self.log('all workers done', debug=True)



from __future__ import annotations

import sys
from math import log10
from typing import Optional, Iterable, TypeVar, List, Tuple, Dict, Any, Type

from django import db
from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.contrib.auth.models import BaseUserManager as DjangoBaseUserManager
from django.db import models, DatabaseError, transaction
from django.db.models.expressions import RawSQL, F
from django.db.transaction import Atomic
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from rest_base.fields import UniqueRandomPositiveInt32, UniqueRandomChar, initialize_base_fields
from rest_base.settings import base_settings

__all__ = [
    'BaseQuerySet', 'BaseManager',
    'BulkUpdateQuerySet', 'BulkUpdateManager',
    'BaseModel', 'AlreadyLocked', 'semaphore',
    'BaseUserManager', 'BaseUser', 'BaseToken',
]

Instance = TypeVar('Instance', bound=models.Model)


class BaseQuerySet(models.QuerySet):
    def update_or_create(self, defaults: Dict[str, Any] = None, **kwargs):
        defaults = defaults or dict()
        self._for_write = True
        with transaction.atomic(using=self.db):
            try:
                obj = self.select_for_update().get(**kwargs)
            except self.model.DoesNotExist:
                params = self._extract_model_params(defaults, **kwargs)
                obj, created = self._create_object_from_params(kwargs, params, lock=True)
                if created:
                    return obj, created
            modified = set()
            for k, v in defaults.items():
                try:
                    old_v = getattr(obj, k)
                except AttributeError:
                    continue
                v = v() if callable(v) else v
                if old_v != v:
                    setattr(obj, k, v)
                    modified.add(k)
            if modified:
                obj.save(update_fields=modified, using=self.db)
        return obj, False


class BaseManager(models.Manager):
    def get_queryset(self):
        return BaseQuerySet(self.model, using=self._db, hints=self._hints)


class BulkUpdateQuerySet(models.QuerySet):
    def bulk_create(
            self, objs: Iterable[Instance],
            batch_size: Optional[int] = base_settings.DB_BATCH_SIZE, ignore_conflicts: bool = False
    ) -> List[Instance]:
        self._for_write = True
        return models.QuerySet(self.model, using=self.db).bulk_create(objs, batch_size, ignore_conflicts)

    def update(self, **kwargs) -> int:
        self._for_write = True
        select_query_set = self.order_by('pk').values('pk')
        return models.QuerySet(self.model, using=self.db).filter(pk__in=RawSQL(
            str(select_query_set.query) + ' FOR NO KEY UPDATE', ()
        )).update(**kwargs)
    update.alters_data = True

    def bulk_update(
            self, objs: Iterable[Instance], fields: Iterable[str],
            batch_size: Optional[int] = base_settings.DB_BATCH_SIZE
    ):
        if not objs:
            return

        self._for_write = True
        pks = [obj.pk for obj in objs]
        query_set = self.filter(pk__in=pks).order_by('pk')
        with transaction.atomic():
            with db.connection.cursor() as cursor:
                cursor.execute(str(query_set.query) + ' FOR NO KEY UPDATE', ())
            return models.QuerySet(self.model, using=self.db).bulk_update(objs, fields, batch_size)
    bulk_update.alters_data = True

    def delete(self) -> Tuple[int, Dict[str, int]]:
        self._for_write = True
        query_set = self.order_by('pk').values('pk')
        return models.QuerySet(self.model, using=self.db).filter(pk__in=RawSQL(
            str(query_set.query) + ' FOR UPDATE', ()
        )).delete()
    delete.alters_data = True
    delete.queryset_only = True


class BulkUpdateManager(models.Manager):
    def get_queryset(self):
        return BulkUpdateQuerySet(self.model, using=self._db, hints=self._hints)


class BaseModelMeta(models.base.ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        abstract = getattr(attrs.get('Meta', None), 'abstract', False)
        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        if not abstract:
            initialize_base_fields(new_class)
        return new_class


class BaseModel(models.Model, metaclass=BaseModelMeta):
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    last_modified = models.DateTimeField(auto_now=True, db_index=True)

    objects = BaseManager()
    bulk_manager = BulkUpdateManager()

    class Meta:
        abstract = True

    def save(
            self, force_insert: bool = False, force_update: bool = False, using: str = None,
            update_fields: Iterable[str] = None
    ):
        if not self._state.adding and update_fields is not None:
            if type(update_fields) is not list:
                update_fields = list(update_fields)
            update_fields.append('last_modified')
        super().save(force_insert, force_update, using, update_fields)
    save.alters_data = True


class AlreadyLocked(Exception):
    def __init__(self, instance):
        self.instance = instance

    def __str__(self):
        return f'AlreadyLocked ({self.instance})'


def semaphore(*, block: bool):
    nowait = not block

    def decorator(cls: Type[models.Model]):
        class Model(cls):
            def __enter__(self: models.Model):
                self._atomic: Atomic = transaction.atomic()
                self._atomic.__enter__()
                try:
                    self.__class__.objects.select_for_update(nowait=nowait).get(pk=self.pk)
                except DatabaseError:
                    self._atomic.__exit__(*sys.exc_info())
                    raise AlreadyLocked(self)

            def __exit__(self, exc_type, exc_val, exc_tb):
                return self._atomic.__exit__(exc_type, exc_val, exc_tb)

        return Model

    return decorator


class BaseUserManager(BaseManager, DjangoBaseUserManager):
    def create_user(self, username: str, password: str):
        user = self.model(username=username)
        user.set_password(password)
        user.save()
        return user

    def create_superuser(self, username: str, password: str):
        user = self.create_user(username, password)
        user.is_staff = True
        user.is_superuser = True
        user.save()
        return user


class BaseUser(BaseModel, AbstractBaseUser, PermissionsMixin):
    user_id = models.PositiveIntegerField(primary_key=True, default=UniqueRandomPositiveInt32)
    username = models.CharField(max_length=base_settings.USERNAME_LENGTH_MAX, unique=True, default=None, null=True, blank=True)
    password = models.CharField(_('password'), max_length=base_settings.PASSWORD_LENGTH_MAX, default='', blank=True)

    is_active = models.BooleanField(default=True)
    is_staff = models.BooleanField(default=False)
    is_guest = models.BooleanField(default=False)

    USERNAME_FIELD = 'username'

    objects = BaseUserManager()

    class Meta:
        abstract = True

    class DeactivatedUser(Exception):
        def __init__(self, user: BaseUser):
            self._user = user

    def __str__(self):
        return self.username

    def raise_for_deactivation(self):
        if not self.is_active:
            raise BaseUser.DeactivatedUser(self)


class BaseToken(BaseModel):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='tokens')
    public_key = models.CharField(max_length=40, unique=True, default=UniqueRandomChar)
    secret_key = models.CharField(max_length=40, unique=True, default=UniqueRandomChar)
    duration = models.DurationField()
    nonce = models.CharField(
        max_length=int(log10(base_settings.NONCE_MAX) + 1) * base_settings.NONCE_WINDOW_SIZE, default='', blank=True)

    class Meta:
        abstract = True

    class TokenMeta:
        duration = base_settings.DEFAULT_TOKEN_DURATION

    class NonceError(Exception):
        pass

    def __str__(self):
        return f'{self.__class__.__name__} ({self.user})'

    @classmethod
    def new(cls, user: BaseUser, duration: timezone.timedelta = None, **kwargs) -> BaseToken:
        user.raise_for_deactivation()

        if duration is None:
            duration = cls.TokenMeta.duration
        token = cls.objects.create(user=user, duration=duration, **kwargs)

        return token

    @classmethod
    def get(cls, public_key: str, next_nonce: int = None, include_user: bool = True) -> Optional[BaseToken]:
        if not public_key:
            return None

        try:
            if include_user:
                queryset = cls.objects.select_related('user')
            else:
                queryset = cls.objects

            token: cls = queryset.get(
                public_key=public_key,
                last_modified__gt=timezone.now() - F('duration'),
                user__is_active=True,
            )

            if next_nonce is None:
                token.save(update_fields=[])
            else:
                if not 0 <= next_nonce < base_settings.NONCE_MAX:
                    return None

                if not token.nonce:
                    token.nonce = str(next_nonce)
                else:
                    nonce_list = [int(n) for n in token.nonce.split(';')]

                    if next_nonce in nonce_list:
                        return None
                    if next_nonce < nonce_list[0]:
                        return None

                    nonce_list.append(next_nonce)
                    nonce_list = sorted(nonce_list)[-base_settings.NONCE_WINDOW_SIZE:]

                    token.nonce = ';'.join(str(n) for n in nonce_list)

                token.save(update_fields=['nonce'])

            return token
        except cls.DoesNotExist:
            return None
